﻿/********************************
 * 
 * autor: Marcin Pietrzykowski
 * mail: mpietrzykowski@wi.zut.edu.pl
 * 
 * ******************************/

using System;
using System.Text;

namespace MatlabUtils {
    /// <summary>
    /// Klasa pozwalająca na korzystanie z instancji Matlaba w każdym jezyku z platformy .NET
    /// Zawiera głownie funkcje tworzące różnego rodzaju wykresy. Ale możliwe jest również wywowłanie dowlnych funkcji.
    /// </summary>
    public sealed class MatlabInstance : IDisposable {
  
        #region Consts
        
        public const string VAR_X_NAME = "x_MA_reserved";

        public const string VAR_Y_NAME = "y_MA_reserved";

        public const string VAR_Z_NAME = "z_MA_reserved";

        public const string WORKSPACE = "base";

        #endregion //end Consts

        #region Fields

        private StringBuilder builder;

        private bool disposed;

        private MLApp.MLApp matlab;

        #endregion //end Fields

        #region Constructors & Destructor

        public MatlabInstance() : this(ServerMode.Single) { }

        public MatlabInstance(ServerMode mode) {
            //to jest standardowe wywołanie matlaba. Nie wiem jak wygląda zwalnianie pamięci w tym wywołaniu dlatego preferowana jest
            //tworzenie obiektu COM i następnie zwalnienie go w metodzie Dispose
            //this.matlab = new MLApp.MLApp();  

            if (mode == ServerMode.Shared) {
                Type matlab = Type.GetTypeFromProgID("Matlab.Application");
                this.matlab = (MLApp.MLApp)Activator.CreateInstance(matlab);
            }
            else {
                Type matlab = Type.GetTypeFromProgID("Matlab.Application.Single");
                this.matlab = (MLApp.MLApp)Activator.CreateInstance(matlab);
            }

            this.disposed = false;            

            this.builder = new StringBuilder();
        }

        ~MatlabInstance() {
            Dispose();
        }

        #endregion //end Constructors & Destructor

        #region Private Methods

        private string addContourLevel(uint levels, string lineSpec) {
            if (levels != 0) {
                if (string.IsNullOrWhiteSpace(lineSpec)) {
                    return levels.ToString();
                }
                else {
                    this.builder.Clear();
                    this.builder.AppendFormat("{0}, {1}", levels, lineSpec);
                    return this.builder.ToString();
                }
            }

            return lineSpec;
        }

        private void addPropertiesToBuilder(string properties) {
            if (!string.IsNullOrWhiteSpace(properties)) {
                this.builder.Length -= 2;
                this.builder.AppendFormat(",{0});", properties);
            }
        }

        private string figureText(string text, string function, string properties = null) {
            this.builder.Clear();
            this.builder.AppendFormat("{0}('{1}');", function, text);
            addPropertiesToBuilder(properties);
            return matlab.Execute(this.builder.ToString());
        }

        #endregion //end Private Methods

        #region Public Methods

        public string Area(double[] x, double[] y, string properties = null) {
            return Execute("area", x, y, properties);
        }

        public string Bar(double[] x, string barColor = null) {
            return Execute("bar", x, barColor);
        }

        public string Compass(double[] x, double[] xi, string lineSpec = null) {
            this.matlab.PutFullMatrix(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x, xi);

            this.builder.Clear();
            this.builder.AppendFormat("compass(" + VAR_X_NAME + ");");

            addPropertiesToBuilder(lineSpec);

            return this.matlab.Execute(this.builder.ToString());
        }

        public string CloseFiguresAndClearAll() {
            return this.matlab.Execute("clear all; close all; clc;");
        }

        public string ClearFigure() {
            return matlab.Execute("clf;");
        }

        public string ClearFigure(uint figureNumber) {
            this.builder.Clear();
            this.builder.AppendFormat("clf({0});", figureNumber);
            return matlab.Execute(this.builder.ToString());
        }

        public string Contour(double[,] x, double[,] y, double[,] z, string lineSpec = null) {
            return Execute("contour", x, y, z, lineSpec);
        }

        public string Contour(double[,] x, double[,] y, double[,] z, uint levels, string lineSpec = null) {
            return Execute("contour", x, y, z, addContourLevel(levels, lineSpec));
        }

        public string ContourFilled(double[,] x, double[,] y, double[,] z, string lineSpec = null) {
            return Execute("contourf", x, y, z, lineSpec);
        }

        public string ContourFilled(double[,] x, double[,] y, double[,] z, uint levels, string lineSpec = null) {
            return Execute("contourf", x, y, z, addContourLevel(levels, lineSpec));
        }

        public string Execute(string text) {
            return this.matlab.Execute(text);
        }

        public string Execute(string function, double[] x, string properties) {
            this.builder.Clear();
            this.builder.AppendFormat("{0}(" + MatlabInstance.VAR_X_NAME + ");", function);

            addPropertiesToBuilder(properties);

            this.matlab.PutWorkspaceData(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x);

            return this.matlab.Execute(this.builder.ToString());
        }

        public string Execute(string function, double[] x, double[] y, string properties, bool checkArrays = true) {
            if (checkArrays && x.Length != y.Length && (x.Length != 1 || y.Length != 1)) {
                throw new ApplicationException("Arrays length have to be equal or at least one array have to be a scalar.");
            }

            this.builder.Clear();
            this.builder.AppendFormat("{0}(" + MatlabInstance.VAR_X_NAME + ", " + MatlabInstance.VAR_Y_NAME + ");", function);

            addPropertiesToBuilder(properties);

            this.matlab.PutWorkspaceData(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x);
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_Y_NAME, MatlabInstance.WORKSPACE, y);

            return this.matlab.Execute(this.builder.ToString());
        }

        public string Execute(string function, double[] x, double[] y, double[] z, string properties) {
            if (x.Length != y.Length || (x.Length != z.Length)) {
                throw new ApplicationException("Arrays length have to be equal.");
            }

            this.builder.Clear();
            this.builder.AppendFormat(
                "{0}(" + MatlabInstance.VAR_X_NAME + ", " + MatlabInstance.VAR_Y_NAME + ", " +  MatlabInstance.VAR_Z_NAME + ");", 
                function
            );

            addPropertiesToBuilder(properties);

            this.matlab.PutWorkspaceData(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x);
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_Y_NAME, MatlabInstance.WORKSPACE, y);
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_Z_NAME, MatlabInstance.WORKSPACE, z);

            return this.matlab.Execute(this.builder.ToString());
        }

        public string Execute(string function, double[,] x, double[,] y, double[,] z, string properties) {
            if (x.GetLength(0) != y.GetLength(0) || (x.GetLength(0) != z.GetLength(0)) ||
                x.GetLength(1) != y.GetLength(1) || (x.GetLength(1) != z.GetLength(1))) {
                throw new ApplicationException("Arrays lengths have to be equal.");
            }

            this.builder.Clear();
            this.builder.AppendFormat(
               "{0}(" + MatlabInstance.VAR_X_NAME + ", " + MatlabInstance.VAR_Y_NAME + ", " + MatlabInstance.VAR_Z_NAME + ");",
               function
           );

            addPropertiesToBuilder(properties);

            this.matlab.PutWorkspaceData(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x);
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_Y_NAME, MatlabInstance.WORKSPACE, y);
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_Z_NAME, MatlabInstance.WORKSPACE, z);

            return this.matlab.Execute(this.builder.ToString());
        }

        public string Figure(uint figureNumber) {
            this.builder.Clear();
            this.builder.AppendFormat("figure({0});", figureNumber);
            return matlab.Execute(this.builder.ToString());
        }

        public string Fill(double[] x, double[] y, string properties = "'b'") {
            return Execute("fill", x, y, properties);
        }

        public string Fill(double[] x, double[] y, double[] z, string properties = "'b'") {
            return Execute("fill3", x, y, z, properties);
        }

        public string GetCharArray(string name, string workspace = MatlabInstance.WORKSPACE) {
            return this.matlab.GetCharArray(name, workspace);
        }

        public void GetFullMatrix(string name, ref Array pr, ref Array pi, string workspace = MatlabInstance.WORKSPACE) {
            this.matlab.GetFullMatrix(name, workspace, ref pr, ref pi);
        }

        public void GetWorkspaceData(string name, out object data, string workspace = MatlabInstance.WORKSPACE) {
            this.matlab.GetWorkspaceData(name, workspace, out data);           
        }

        public string GridOn() {           
            return matlab.Execute("grid on;");
        }

        public string GridOff() {
            return matlab.Execute("grid off;");
        }

        public string Hist(double[] x, uint numberOfBins = 10) {
            return Execute("hist", x, numberOfBins.ToString());
        }

        public string Hist(double[] x, double[] centerOfIntervals) {
            return Execute("hist", x, centerOfIntervals, null, false);
        }

        public string HoldOn() {
            return matlab.Execute("hold on;");
        }

        public string HoldOff() {
            return matlab.Execute("hold off;");
        }

        public string Mesh(double[,] x, double[,] y, double[,] z, string properties = null) {
            return Execute("mesh", x, y, z, properties);
        }

        public string MeshContour(double[,] x, double[,] y, double[,] z) {
            return Execute("meshc", x, y, z, null);
        }

        public string Plot(double[] x, double[] y, string lineSpec = null) {
            return Execute("plot", x, y, lineSpec);
        }

        public string Plot(double[] x, double[] y, double[] z, string lineSpec = null) {
            return Execute("plot3", x, y, z, lineSpec);
        }

        public string Points(double[] x, double[] y, string lineSpec = "'.'") {
            return Execute("plot", x, y, lineSpec);
        }

        public string Points(double[] x, double[] y, double[] z, string lineSpec = "'.'") {
            return Execute("plot3", x, y, z, lineSpec);
        }

        public string Polar(double[] angles, double[] radiuses, string lineSpec = null) {
            return Execute("polar", angles, radiuses, lineSpec);
        }

        public string PolarFromCartesianCoordinates(double[] x, double[] y, double centreX = 0.0, double centreY = 0.0, string lineSpec = null) {
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x);
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_Y_NAME, MatlabInstance.WORKSPACE, y);

            this.builder.Clear();
            this.builder.AppendFormat(
                MatlabInstance.VAR_X_NAME + " = " + MatlabInstance.VAR_X_NAME + " - {0};" +
                MatlabInstance.VAR_Y_NAME + " = " + MatlabInstance.VAR_Y_NAME + " - {1};" +
                "[" + MatlabInstance.VAR_X_NAME + ", " + MatlabInstance.VAR_Y_NAME + "] = cart2pol(" + MatlabInstance.VAR_X_NAME + ", " + MatlabInstance.VAR_Y_NAME + ");"+
                "polar(" + MatlabInstance.VAR_X_NAME + ", " + MatlabInstance.VAR_Y_NAME + ");",
                centreX.ToString(System.Globalization.CultureInfo.InvariantCulture),
                centreY.ToString(System.Globalization.CultureInfo.InvariantCulture)
            );

            addPropertiesToBuilder(lineSpec);

            return this.matlab.Execute(this.builder.ToString());
        }

        public void PutCharArray(string name, string charArray, string workspace = MatlabInstance.WORKSPACE) {
            this.matlab.PutCharArray(name, workspace, charArray);
        }

        public void PutFullMatrix(string name, Array pr, Array pi, string workspace = MatlabInstance.WORKSPACE) {
            this.matlab.PutFullMatrix(name, workspace, pr, pi);
        }

        public void PutWorkspaceData(string name, object data, string workspace = MatlabInstance.WORKSPACE) {
            this.matlab.PutWorkspaceData(name, workspace, data);
        }

        public string Ribbon(double[,] x) {
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x);
            return this.matlab.Execute("ribbon(" + MatlabInstance.VAR_X_NAME + ");");
        }

        public string Ribbon(double[,] x, double[,] y, double width = 0.75) {
            if (x.GetLength(0) != y.GetLength(0) || x.GetLength(1) != y.GetLength(1)) {
                throw new ApplicationException("Arrays lengths have to be equal.");
            }

            this.builder.Clear();
            this.builder.AppendFormat("ribbon(" + MatlabInstance.VAR_X_NAME + ", " + MatlabInstance.VAR_Y_NAME + ",{0});", width.ToString(System.Globalization.CultureInfo.InvariantCulture));

            this.matlab.PutWorkspaceData(MatlabInstance.VAR_X_NAME, MatlabInstance.WORKSPACE, x);
            this.matlab.PutWorkspaceData(MatlabInstance.VAR_Y_NAME, MatlabInstance.WORKSPACE, y);

            return this.matlab.Execute(this.builder.ToString());
        }

        public string Rose(double[] angles, uint nbins = 20) {
            return Execute("rose", angles, nbins.ToString());
        }

        public string SaveFigureObject(uint figureNumber, string fileName) {
            this.builder.Clear();
            this.builder.AppendFormat("savefig({0}, {1})", figureNumber, fileName);
            return this.matlab.Execute(this.builder.ToString());
        }

        public string SaveFigurePlot(uint figureNumber, string filename, string resolution = "-r300", string filetype = "-dpng") {
            this.builder.Clear();
            this.builder.AppendFormat("print({0}, '{1}', '{2}', '{3}')", figureNumber, filetype, resolution, filename);
            return this.matlab.Execute(this.builder.ToString());
        }

        public string Surf(double[,] x, double[,] y, double[,] z, string properties = null) {
            return Execute("surf", x, y, z, properties);
        }

        public string SurfContour(double[,] x, double[,] y, double[,] z, string properties = null) {
            return Execute("surfc", x, y, z, properties);
        }

        public string Title(string text, string properties = null) {
            return figureText(text, "title", properties);
        }

        public string Text(double x, double y, string text, string properties = null) {
            this.builder.Clear();
            this.builder.AppendFormat(
                "text({0}, {1}, '{2}');",
                x.ToString(System.Globalization.CultureInfo.InvariantCulture), 
                y.ToString(System.Globalization.CultureInfo.InvariantCulture),
                text
            );
            addPropertiesToBuilder(properties);
            return this.matlab.Execute(this.builder.ToString());
        }

        public string Waterfall(double[,] x, double[,] y, double[,] z) {
            return Execute("waterfall", x, y, z, null);
        }

        public string XLabel(string text, string properties = null) {
            return figureText(text, "xlabel", properties);
        }

        public string YLabel(string text, string properties = null) {
            return figureText(text, "ylabel", properties);
        }

        public string ZLabel(string text, string properties = null) {
            return figureText(text, "zlabel", properties);
        }

        #endregion //end Public Methods

        #region IDisposable Members

        public void Dispose() {
            if (!this.disposed) {
                System.Runtime.InteropServices.Marshal.ReleaseComObject(matlab);
                this.disposed = true;
            }

            GC.SuppressFinalize(this);
        }

        #endregion //end IDisposable Members
    }
}
